from PIL import Image
from torch.utils.data import Dataset
from .format import Conversation, ConverstationText, ConverstationImage, DatasetOutput
import json
import os


class ChemDataset(Dataset):
    def __init__(
        self, vis_root, ann_path, vis_processor=None, text_processor=None, split="train"
    ):
        """
        vis_root (string): Root directory of images (e.g. coco/images/)
        ann_root (string): directory to store the annotation file
        """
        self.vis_root = vis_root

        self.vis_processor = (
            vis_processor if vis_processor is not None else self._vis_processor
        )
        self.text_processor = text_processor
        if split == "train":
            self.data = self.create_data(ann_path, split="train")
        elif split == "test":
            self.data = self.create_data(ann_path, split="test")

    def _vis_processor(self, image: Image.Image):
        width, height = image.size
        if width > 500 or height > 500:
            max_size = max(width, height)
            ratio = 500 / max_size
            new_width = int(width * ratio)
            new_height = int(height * ratio)
            image = image.resize((new_width, new_height), Image.Resampling.BILINEAR)

        if width < 28 or height < 28:
            min_size = min(width, height)
            ratio = 28 / min_size + 1
            new_width = int(width * ratio)
            new_height = int(height * ratio)
            image = image.resize((new_width, new_height), Image.Resampling.BILINEAR)

        return image

    def create_data(self, ann_path, split=1):
        import os.path as osp

        if split == "train":
            json_path = osp.join(ann_path, "conversations_loc_train.jsonl")
        elif split == "test":
            json_path = osp.join(ann_path, "conversations_loc_val.jsonl")
        processed_data = []
        with open(json_path, "r") as f:
            for line in f:
                data = json.loads(line)
                image_file = osp.join(self.vis_root, data["images"][0].split("/")[-1])
                conversation = data["messages"]
                system = conversation[0]["content"]
                query = conversation[1]["content"]
                answer = conversation[2]["content"]
                processed_data.append(
                    {
                        "question": query,
                        "answer": answer,
                        "image_path": image_file,
                        "system": system,
                        "original": data,
                    }
                )
        return processed_data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        sample = self.data[index]
        image = Image.open(os.path.join(self.vis_root, sample["image_path"])).convert(
            "RGB"
        )
        question = sample["question"]
        answer = sample["answer"]
        if self.vis_processor is not None:
            image = self.vis_processor(image)
        if self.text_processor is not None:
            question = self.text_processor(question)
            answer = self.text_processor(answer)

        chat = [
            Conversation(
                type="system",
                content=[ConverstationText(type="text", text=sample["system"])],
            ),
            Conversation(
                type="user",
                content=[
                    ConverstationImage(type="image", image_url=""),
                    ConverstationText(
                        type="text", text=f"[vqa] {question.replace('<image>','')}"
                    ),
                ],
            ),
            Conversation(
                type="assistant", content=[ConverstationText(type="text", text=answer)]
            ),
        ]

        return DatasetOutput(
            images=[image],
            chat=chat,
            answer=answer,
            original=sample["original"],
        )


class ChemDatasetForGeneration(ChemDataset):
    def __getitem__(self, index):
        sample = self.data[index]
        image = Image.open(os.path.join(self.vis_root, sample["image_path"])).convert(
            "RGB"
        )
        question = sample["question"]
        answer = sample["answer"]
        if self.vis_processor is not None:
            image = self.vis_processor(image)
        if self.text_processor is not None:
            question = self.text_processor(question)
            answer = self.text_processor(answer)

        chat = [
            Conversation(
                type="system",
                content=[ConverstationText(type="text", text=sample["system"])],
            ),
            Conversation(
                type="user",
                content=[
                    ConverstationImage(type="image", image_url=""),
                    ConverstationText(
                        type="text", text=f"[vqa] {question.replace('<image>','')}"
                    ),
                ],
            ),
        ]

        return DatasetOutput(
            images=[image],
            chat=chat,
            answer=answer,
            original=sample["original"],
        )
