from PIL import Image
from .format import (
    Conversation,
    ConverstationAudio,
    ConverstationImage,
    ConverstationText,
    DatasetOutput,
)
from torch.utils.data import Dataset
import json
import os
from pathlib import Path


class OCRVQADataset(Dataset):
    def __init__(
        self, vis_root, ann_path, vis_processor=None, text_processor=None, split="train"
    ):
        """
        vis_root (string): Root directory of images (e.g. coco/images/)
        ann_root (string): directory to store the annotation file
        """
        self.vis_root = vis_root

        from .vis_processor import size_processor

        self.vis_processor = (
            vis_processor if vis_processor is not None else size_processor
        )
        self.text_processor = text_processor
        self.split = split
        self.ann_path = ann_path

    def load_data(self):
        if self.split == "train":
            self.data = self.create_data(Path(self.ann_path, "dataset.json"), split=1)
        elif self.split == "test":
            self.data = self.create_data(Path(self.ann_path, "dataset.json"), split=3)

        # self.instruction_pool = [
        #     "[vqa] {}",
        #     "[vqa] Based on the image, respond to this question with a short answer: {}",
        # ]

    def create_data(self, ann_path, split=1):
        processed_data = []
        with open(ann_path, "r") as f:
            data = json.load(f)
        for k in data.keys():
            if data[k]["split"] != split:
                continue  # 1 for training, 2 for validation, 3 for test
            ext = os.path.splitext(data[k]["imageURL"])[1]
            imageFile = k + ext
            assert len(data[k]["questions"]) == len(data[k]["answers"])
            for q, a in zip(data[k]["questions"], data[k]["answers"]):
                if os.path.exists(os.path.join(self.vis_root, imageFile)):
                    processed_data.append(
                        {
                            "question": q,
                            "answer": a,
                            "image_path": imageFile,
                            "image_id": k,
                            "title": data[k]["title"],
                            "genre": data[k]["genre"],
                            "original": data[k],
                        }
                    )
        return processed_data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        sample = self.data[index]
        image: Image.Image = Image.open(
            os.path.join(self.vis_root, sample["image_path"])
        ).convert("RGB")
        # resize image

        question = sample["question"]
        answer = sample["answer"]
        if self.vis_processor is not None:
            image = self.vis_processor(image)
        if self.text_processor is not None:
            question = self.text_processor(question)
            answer = self.text_processor(answer)

        chat = [
            Conversation(
                role="user",
                content=[
                    ConverstationImage(type="image", image_url=""),
                    ConverstationText(
                        type="text",
                        text=f"[vqa] Based on the image, respond to this question with a short answer: {question}",
                    ),
                ],
            ),
            Conversation(
                role="assistant", content=[ConverstationText(type="text", text=answer)]
            ),
        ]

        return DatasetOutput(
            chat=chat,
            original=sample["original"],
            images=[image],
        )  # type: ignore


class OCRVQADatasetForGeneration(OCRVQADataset):

    def __getitem__(self, index):
        sample = self.data[index]
        image = Image.open(os.path.join(self.vis_root, sample["image_path"])).convert(
            "RGB"
        )
        # resize image
        question = sample["question"]
        answer = sample["answer"]
        if self.vis_processor is not None:
            image = self.vis_processor(image)
        if self.text_processor is not None:
            question = self.text_processor(question)
            answer = self.text_processor(answer)

        chat = [
            Conversation(
                role="user",
                content=[
                    ConverstationImage(type="image", image_url=""),
                    ConverstationText(
                        type="text",
                        text=f"[vqa] Based on the image, respond to this question with a short answer: {question}",
                    ),
                ],
            ),
        ]
        return DatasetOutput(
            images=[image],
            chat=chat,
            answer=answer,
            original=sample["original"],
        )  # type: ignore


from .factory import register_dataset
from .format import dataset_dir as base_path

dataset = {
    "train": OCRVQADataset(
        vis_root=Path(base_path, "OCR-VQA-200K", "images"),
        ann_path=Path(base_path, "OCR-VQA-200K"),
        split="train",
    ),
    "test": OCRVQADataset(
        vis_root=Path(base_path, "OCR-VQA-200K", "images"),
        ann_path=Path(base_path, "OCR-VQA-200K"),
        split="test",
    ),
    "generation": OCRVQADatasetForGeneration(
        vis_root=Path(base_path, "OCR-VQA-200K", "images"),
        ann_path=Path(base_path, "OCR-VQA-200K"),
        split="test",
    ),
}

register_dataset("ocrvqa200k", dataset, tag=["image", "text"])
