from .format import (
    Conversation,
    ConverstationAudio,
    ConverstationImage,
    ConverstationText,
    DatasetOutput,
)
from torch.utils.data import Dataset
from datasets import load_dataset, DatasetDict
from typing import Literal


class RefCOCOplusDataset(Dataset):
    def __init__(
        self,
        vis_processor=None,
        text_processor=None,
        split: Literal["val", "testA"] = "val",
    ):
        """
        vis_root (string): Root directory of images (e.g. coco/images/)
        ann_root (string): directory to store the annotation file
        """

        self.vis_processor = vis_processor
        self.text_processor = text_processor
        self.split = split

    def load_data(self):
        from .format import dataset_dir

        ds = load_dataset("lmms-lab/RefCOCOplus", cache_dir=dataset_dir)  # type: ignore
        self.data = ds[self.split]  # type: ignore

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        sample = self.data[index]
        # print(sample)
        images = sample["image"]
        question = sample["question"]
        answer = sample["answer"]

        if self.vis_processor is not None:
            images = self.vis_processor(images)
        if self.text_processor is not None:
            question = self.text_processor(question)

        chat = [
            Conversation(
                role="user",
                content=[
                    ConverstationImage(type="image", image_url=""),
                    ConverstationText(
                        type="text",
                        text=question,
                    ),
                ],
            ),
            Conversation(
                role="assistant",
                content=[ConverstationText(type="text", text=answer)],
            ),
        ]

        return DatasetOutput(
            images=[images],
            chat=chat,
            original=sample,
        )  # type: ignore


class RefCOCOplusDatasetForGeneration(RefCOCOplusDataset):

    def __getitem__(self, index):
        sample = self.data[index]
        # print(sample)
        images = sample["image"]
        question = sample["question"]
        answer = sample["answer"]

        if self.vis_processor is not None:
            images = self.vis_processor(images)
        if self.text_processor is not None:
            question = self.text_processor(question)

        chat = [
            Conversation(
                role="user",
                content=[
                    ConverstationImage(type="image", image_url=""),
                    ConverstationText(
                        type="text",
                        text=f"{question}",
                    ),
                ],
            ),
        ]

        return DatasetOutput(
            images=[images],
            chat=chat,
            answer=answer,
            original=sample,
        )  # type: ignore


def test_RefCOCOplus():
    dataset = RefCOCOplusDataset(
        split="val",
    )
    print(dataset[3])
    assert len(dataset) > 0
    assert len(dataset[0]["chat"]) > 0
    dataset = RefCOCOplusDatasetForGeneration(
        split="testA",
    )
    print(dataset[3])
    assert len(dataset) > 0
    assert len(dataset[0]["chat"]) > 0


dataset = {
    "train": RefCOCOplusDataset(split="val"),
    "test": RefCOCOplusDataset(split="testA"),
    "generation": RefCOCOplusDatasetForGeneration(split="testA"),
}
from .factory import register_dataset

register_dataset(
    dataset_name="refcocoplus",
    dataset=dataset,
    tag=["image", "text"],
)


if __name__ == "__main__":
    test_RefCOCOplus()
