from torch.utils.data import Dataset
from typing import Literal, List
from pathlib import Path
from dataset_library.format import dataset_dir

MAPPING_NAME_TO_DATASET: dict[
    str, dict[Literal["train", "test", "generation"], Dataset]
] = {}


def register_dataset(
    dataset_name: str,
    dataset: dict[Literal["train", "test", "generation"], Dataset],
    tag: List[Literal["image", "text", "audio", "video"]] = [],
    **kwargs,
) -> None:
    """
    Register a dataset.

    Args:
        dataset_name (`str`):
            The name of the dataset.
        dataset (`Dataset`):
            The dataset to register.
    """
    dataset_name = dataset_name.lower()
    if dataset_name in MAPPING_NAME_TO_DATASET:
        raise ValueError(f"Dataset {dataset_name} already registered.")
    MAPPING_NAME_TO_DATASET[dataset_name] = dataset


from .GigaspeechDataset import (
    GigaspeechDataset,
    GigaspeechDatasetForGeneration,
)
from .OCRVQA200KDataset import OCRVQADataset, OCRVQADatasetForGeneration
from .ChemDataset import ChemDataset, ChemDatasetForGeneration
from .TextVQADataset import TextVQADataset, TextVQADatasetForGeneration
from .ScienceQADataset import (
    ScienceQADataset,
    ScienceQADatasetForGeneration,
)
from .RefCOCODataset import (
    RefCOCODataset,
    RefCOCODatasetForGeneration,
)
from .RefCOCOgDataset import (
    RefCOCOgDataset,
    RefCOCOgDatasetForGeneration,
)
from .RefCOCOPlusDataset import (
    RefCOCOplusDataset,
    RefCOCOplusDatasetForGeneration,
)
from .VizWizDataset import (
    VizWizDataset,
    VizWizDatasetForGeneration,
)


def get_dataset(
    dataset_name: str, base_path=dataset_dir
) -> dict[Literal["train", "test", "generation"], Dataset]:
    """
    Get a dataset by name.

    Args:
        dataset_name (`str`):
            The name of the dataset.
        base_path (`str`):
            The base path to the dataset.
    """
    dataset_name = dataset_name.lower()
    if dataset_name not in MAPPING_NAME_TO_DATASET:
        raise ValueError(f"Dataset {dataset_name} not registered.")
    for key in MAPPING_NAME_TO_DATASET[dataset_name]:
        MAPPING_NAME_TO_DATASET[dataset_name][key].load_data()
    return MAPPING_NAME_TO_DATASET[dataset_name]
