# from transformers import Qwen2VLProcessor
import sys

sys.path.insert(0, "transformers_repo/src/")
sys.path.insert(0, "peft_repo/src/")
import transformers
import peft
from dataset_library.format import DatasetOutput

import torch

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from transformers import Qwen2VLProcessor
    from transformers import Qwen2_5_VLProcessor


def collate_fn_for_train(examples: list[DatasetOutput], processor: "Qwen2VLProcessor"):
    # Get the texts and images, and apply the chat template
    texts = [
        processor.apply_chat_template(example["chat"], tokenize=False)
        for example in examples
    ]
    # print(texts)
    images = [
        example["images"] for example in examples if example["images"][0] is not None
    ]
    images = images if len(images) > 0 else None

    # Tokenize the texts and process the images
    batch = processor(text=texts, images=images, return_tensors="pt", padding=True)

    # The labels are the input_ids, and we mask the padding tokens in the loss computation
    labels = batch["input_ids"].clone()
    labels[labels == processor.tokenizer.pad_token_id] = -100  #
    # Ignore the image token index in the loss computation (model specific)
    # 对<|im_start|>system *** <|im_end|>\n加掩码
    im_start_token_id = processor.tokenizer.convert_tokens_to_ids("<|im_start|>")
    im_end_token_id = processor.tokenizer.convert_tokens_to_ids("<|im_end|>")
    system_token_id = processor.tokenizer.convert_tokens_to_ids("system")
    user_token_id = processor.tokenizer.convert_tokens_to_ids("user")
    assistant_token_id = processor.tokenizer.convert_tokens_to_ids("assistant")
    # enter_token_id = processor.tokenizer.convert_tokens_to_ids("\n")
    # print(im_start_token_id, im_end_token_id, system_token_id, user_token_id, assistant_token_id, enter_token_id, processor.tokenizer.pad_token_id)
    # 151644 151645 8948 872 77091 None 151643

    for i, label in enumerate(labels):
        now_index = 0
        while now_index < len(label):
            if label[now_index] == im_start_token_id:
                label[now_index] = -100
                now_index += 1
                if (
                    label[now_index] == system_token_id
                    or label[now_index] == user_token_id
                ):
                    while label[now_index] != im_end_token_id:
                        label[now_index] = -100
                        now_index += 1
                    label[now_index] = -100
                elif label[now_index] == assistant_token_id:
                    label[now_index] = -100
                    label[now_index + 1] = -100
                    now_index += 2
                    while (
                        now_index < len(label) and label[now_index] != im_end_token_id
                    ):
                        now_index += 1
            now_index += 1
    batch["labels"] = labels
    # batch["task_id"] = torch.tensor([0] * len(labels), dtype=torch.long)

    return batch


def collate_fn_for_evaluate(examples, processor: "Qwen2VLProcessor"):
    # Get the texts and images, and apply the chat template
    texts = [
        processor.apply_chat_template(
            example["chat"], tokenize=False, add_generation_prompt=True
        )
        for example in examples
    ]
    # print(texts)
    images = [
        example["images"] for example in examples if example["images"][0] is not None
    ]
    images = images if len(images) > 0 else None

    # Tokenize the texts and process the images
    batch = processor(text=texts, images=images, return_tensors="pt", padding=True)

    answers = [example["answer"] for example in examples]
    answers = processor(text=answers, return_tensors="pt", padding=True)
    batch["answers_ids"] = answers["input_ids"]
    batch["original_data"] = [example["original"] for example in examples]
    # input_ids torch.Size([3, 370])
    # attention_mask torch.Size([3, 370])
    # pixel_values torch.Size([3888, 1176])
    # image_grid_thw torch.Size([3, 3])
    # answers_ids torch.Size([3, 10])
    # answers_mask torch.Size([3, 10])
    return batch


if __name__ == "__main__":

    from transformers import AutoProcessor

    processor = AutoProcessor.from_pretrained("Qwen/Qwen2.5-VL-3B-Instruct")
    from PIL import Image

    # 随机生成一个图片
    import numpy as np

    random_image = Image.fromarray(
        np.random.randint(0, 255, (100, 100, 3), dtype=np.uint8)
    )
    example = {
        "chat": [
            # {"role": "user", "content": "What is the capital of France?"},
            {
                "role": "user",
                "content": [
                    {
                        "type": "image",
                        "image_url": "",  # Assuming no image for this example
                    },
                    {
                        "type": "image",
                        "image_url": "",  # Assuming no image for this example
                    },
                    {
                        "type": "image",
                        "image_url": "",  # Assuming no image for this example
                    },
                    {
                        "type": "text",
                        "text": "What is the capital of France?",
                    },
                ],
            },  # Assuming no image for this example
            {"role": "assistant", "content": "The capital of France is Paris."},
        ],
        "images": [
            random_image,
            random_image,
            random_image,
        ],  # Assuming no images for this example
    }
    batch = collate_fn_for_train([example], processor)
    # print(batch)
    for k, v in batch.items():
        if isinstance(v, torch.Tensor):
            print(f"{k}: {v.shape}")
        else:
            print(f"{k}: {v}")
    # input_ids: torch.Size([1, 101])
    # attention_mask: torch.Size([1, 101])
    # pixel_values: torch.Size([256, 1176])
    # image_grid_thw: torch.Size([1, 3])
    # labels: torch.Size([1, 101])
    # Load model directly
    from transformers.models.qwen2_5_omni import Qwen2_5OmniThinkerForConditionalGeneration, Qwen2_5OmniProcessor
    model = Qwen2_5OmniThinkerForConditionalGeneration.from_pretrained("Qwen/Qwen2.5-Omni-3B", torch_dtype="auto", device_map="auto")
    processor = Qwen2_5OmniProcessor.from_pretrained("Qwen/Qwen2.5-Omni-3B")


