from . import RegularizationMethod
import torch


class LWF(RegularizationMethod):
    """Learning Without Forgetting.

    The method applies knowledge distilllation to mitigate forgetting.
    The teacher is the model checkpoint after the last experience.
    """

    def __init__(self, LWF_lambda=1, temperature=2):
        """
        :param alpha: distillation hyperparameter. It can be either a float
                number or a list containing alpha for each experience.
        :param temperature: softmax temperature for distillation
        """
        self.LWF_lambda = LWF_lambda
        self.temperature = temperature
        self.previous_logits = {}
        """ In Avalanche, targets of different experiences are not ordered. 
        As a result, some units may be allocated even though their 
        corresponding class has never been seen by the model.
        Knowledge distillation uses only units corresponding
        to old classes. 
        """

    def adapt(self, output, **kwargs):
        def modified_kl_div(old, new):
            return -torch.mean(torch.sum(old * torch.log(new), 1))

        def smooth(logits, temp, dim):
            log = logits ** (1 / temp)
            return log / torch.sum(log, dim).unsqueeze(1)

        lwf_loss = []

        soft = torch.nn.Softmax(dim=1)

        previous_keys = self.previous_logits.keys()

        for index, question_id in enumerate(iterable=kwargs["question_ids"]):
            if question_id in previous_keys:
                previous_logits = self.previous_logits[question_id]
                current_logits = output["logits"][index]
                short_index = min(len(previous_logits), len(current_logits))
                previous_logits = previous_logits[:short_index]
                current_logits = current_logits[:short_index]
                lwf_loss.append(
                    modified_kl_div(
                        old=smooth(
                            logits=soft(previous_logits).to(current_logits.device),
                            temp=2,
                            dim=1,
                        ),
                        new=smooth(logits=soft(current_logits), temp=2, dim=1),
                    )
                )
        if len(lwf_loss) > 0:
            output["loss"] += self.LWF_lambda * torch.stack(
                tensors=lwf_loss, dim=0
            ).sum(dim=0)
        return output

    def update_previous_logits(self, question_id, logits):
        """Update the previous logits for the given question id."""
        self.previous_logits[question_id] = logits
