import evaluate
from accelerate import Accelerator


def evaluate_rouge(model, val_dataloader, processor, accelerator: Accelerator = None):
    glue = evaluate.load("rouge")

    for batch in val_dataloader:
        completion = model.generate(
            input_ids=batch["input_ids"],
            attention_mask=batch["attention_mask"],
            pixel_values=batch["pixel_values"],
            image_grid_thw=batch["image_grid_thw"],
            max_length=1000,
        )
        target = batch["answers_ids"]
        generated_text = [
            out_ids[len(in_ids) :]
            for out_ids, in_ids in zip(completion, batch["input_ids"])
        ]
        generated_text = processor.tokenizer.batch_decode(
            generated_text, skip_special_tokens=True
        )
        target_text = processor.tokenizer.batch_decode(target, skip_special_tokens=True)
        glue.add_batch(predictions=generated_text, references=target_text)

    print(glue.compute())


def evalute_save(model, val_dataloader, processor, accelerator: Accelerator = None):
    import os

    mtime = 0
    for root, dirs, files in os.walk("."):
        for file in files:
            time = os.path.getmtime(os.path.join(root, file))
            if time > mtime:
                mtime = time

    # 获取目录最后修改时间
    if accelerator.is_local_main_process:
        if not os.path.exists(f"results/{mtime}"):
            os.makedirs(f"results/{mtime}")

    from tqdm import tqdm

    if accelerator.is_local_main_process:
        bar = tqdm(total=len(val_dataloader))

    for batch in val_dataloader:
        target = batch.pop("answers_ids")
        origianl = batch.pop("original_data")
        answers = []
        completion = model.generate(
            **batch,
            max_length=1000,
        )
        generated_text = [
            out_ids[len(in_ids) :]
            for out_ids, in_ids in zip(completion, batch["input_ids"])
        ]
        generated_text = processor.tokenizer.batch_decode(
            generated_text, skip_special_tokens=True
        )
        target_text = processor.tokenizer.batch_decode(target, skip_special_tokens=True)
        for i in range(len(generated_text)):
            answers.append(
                {
                    "generated": generated_text[i],
                    "target": target_text[i],
                    "original": str(origianl[i]),
                }
            )
        import json

        world_size = accelerator.process_index

        with open(f"results/{mtime}/answers_{world_size}.jsonl", "a") as f:
            for answer in answers:
                f.write(json.dumps(answer) + "\n")

        if accelerator.is_local_main_process:
            bar.update(1)
    accelerator.wait_for_everyone()
    if accelerator.is_local_main_process:
        bar.close()
        # merge file
        answers = []
        files = [
            file for file in os.listdir(f"results/{mtime}") if file.endswith(".jsonl")
        ]
        for file in files:
            with open(f"results/{mtime}/{file}", "r") as f:
                answers.extend(f.readlines())
        with open(f"results/{mtime}/answers.jsonl", "w") as f:
            f.writelines(answers)
        # delete file
        for file in files:
            os.remove(f"results/{mtime}/{file}")
